@@ -1,7 +1,7 @@
 /*
  * cwdaemon - morse sounding daemon for the parallel or serial port
- * Copyright (C) 2002 -2004 Joop Stakenborg <pg4i@amsat.org>
- *                       and many authors, see the AUTHORS file.
+ * Copyright (C) 2002 -2005 Joop Stakenborg <pg4i@amsat.org>
+ *		       and many authors, see the AUTHORS file.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -87,14 +87,11 @@
 #if (defined(__unix__) || defined(unix)) && !defined(USG)
 # include <sys/param.h>
 #endif
+#include <limits.h>
 
 #include "cwlib.h"
 #include "cwdaemon.h"
 
-#ifdef HAVE_DEV_PPBUS_PPI_H
-static char *actual_device(const char *in);
-#endif
-
 /* network vars */
 int sin_len, reply_socklen;
 int socket_descriptor;
@@ -128,6 +125,8 @@ struct timespec sleeptime, time_remainder; /* delay timers */
 #define MAXMORSE 4000
 char morsetext[MAXMORSE];
 
+char *keydev;
+
 cwdevice cwdevice_ttys = {
 	init:ttys_init,
 	free:ttys_free,
@@ -139,6 +138,17 @@ cwdevice cwdevice_ttys = {
 	footswitch:NULL
 };
 
+cwdevice cwdevice_null = {
+       init:null_init,
+       free:null_free,
+       reset:null_reset,
+       cw:null_cw,
+       ptt:null_ptt,
+       ssbway:NULL,
+       switchband:NULL,
+       footswitch:NULL
+};
+
 #if defined (HAVE_LINUX_PPDEV_H) || defined (HAVE_DEV_PPBUS_PPI_H)
 cwdevice cwdevice_lp = {
 	init:lp_init,
@@ -150,12 +160,8 @@ cwdevice cwdevice_lp = {
 	switchband:lp_switchband,
 	footswitch:lp_footswitch
 };
-
-cwdevice *cwdev = &cwdevice_lp;
-#else
-cwdevice *cwdev = &cwdevice_ttys;
 #endif
-
+cwdevice *cwdev;
 
 /* catch ^C when running in foreground */
 static RETSIGTYPE
@@ -303,19 +309,32 @@ initmorse (void)
 	cw_set_gap (0);
 }
 
+/* properly parse a 'long' integer */
+static int
+get_long(const char *buf, long *lvp)
+{
+	char *ep;
+	long lv;
+
+	errno = 0;
+	lv = strtol(buf, &ep, 10);
+	if (buf[0] == '\0' || *ep != '\0')
+		return (-1);
+	if (errno == ERANGE && (lv == LONG_MAX || lv == LONG_MIN))
+		return (-1);
+	*lvp = lv;
+	return (0);
+}
+
 /* watch the socket and if there is an escape character check what it is,
    otherwise play morse. Return 0 with escape characters and empty messages.*/
 static int
 recv_code (void)
 {
-	char message[257];
+	char message[257], *ndev;
 	ssize_t recv_rc;
-	int speed = 24;
-	int weight = 0;
-	int f = 800;
-	int vol = 70;
-	int valid_sdevice = 0;
-	int tunesec;
+	int valid_sdevice = 0, fd;
+	long lv;
 
 	recv_rc = recvfrom (socket_descriptor, message, sizeof (message) - 1,
 		0, (struct sockaddr *) &k_sin, &sin_len);
@@ -328,9 +347,9 @@ recv_code (void)
 
 	if (recv_rc > 0)
 	{
+		message[recv_rc] = '\0';
 		if (message[0] != 27)
 		{	/* no ESCAPE */
-			message[recv_rc] = '\0';
 			if ((strlen (message) + strlen (morsetext)) <= MAXMORSE - 1)
 				strcat (morsetext, message);
 			return 1;
@@ -352,25 +371,27 @@ recv_code (void)
 				debug ("Reset all values");
 				break;
 			case '2':	/*speed */
-				speed = atoi (message + 2);
-				if (speed > 4 && speed < 61)
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv > 4 && lv < 61)
 				{
-					morse_speed = speed;
+					morse_speed = lv;
 					cw_set_send_speed (morse_speed);
 					debug ("Speed: %d wpm", morse_speed);
 				}
 				break;
 			case '3':	/* tone */
-				f = atoi (message + 2);
-				if (f > 0 && f < 4001)
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv > 0 && lv < 4001)
 				{
-					morse_tone = f;
+					morse_tone = lv;
 					cw_set_frequency (morse_tone);
 					cw_set_volume (70);
 					morse_sound = 1;
 					debug ("Tone: %s Hz, volume 70%", message + 2);
 				}
-				else if (f == 0)	/* sidetone off */
+				else if (lv == 0)	/* sidetone off */
 				{
 					morse_tone = 0;
 					cw_set_volume (0);
@@ -399,57 +420,48 @@ recv_code (void)
 				debug ("Wordmode set");
 				break;
 			case '7':	/* set weighting */
-				weight = atoi (message + 2);
-				if ((weight > -51) && (weight < 51))	/* only allowed range */
+				if (get_long(message + 2, &lv))
+					break;
+				if ((lv > -51) && (lv < 51))	/* only allowed range */
 				{	
-					cw_set_weighting (weight);
-					debug ("Weight: %d", weight);
+					cw_set_weighting (lv);
+					debug ("Weight: %ld", lv);
 				}
 				break;
 			case '8':	/* device type */
-				if (!strncmp (message + 2, "tty", 3))
+				ndev = message + 2;
+				if ((fd = dev_is_tty(ndev)) != -1)
 				{
-					cwdev->free (cwdev);
+					cwdev->free(cwdev);
 					cwdev = &cwdevice_ttys;
-#ifdef BSD
-					if (!strncmp (message + 2, "ttyS", 4))
-						message[3] = 'd';
-#endif
-					cwdev->desc = message + 2;
-					cwdev->init (cwdev);
+					keydev = cwdev->desc = ndev;
+					cwdev->init(cwdev, fd);
 				}
-#ifdef HAVE_LINUX_PPDEV_H
-				else if (!strncmp (message + 2, "parport", 7))
-				{	
-					cwdev->free (cwdev);
-					cwdev = &cwdevice_lp;
-					cwdev->desc = message + 2;
-					cwdev->init (cwdev);
-				}
-#endif
-#ifdef HAVE_DEV_PPBUS_PPI_H
-				else if (strncmp (message + 2, "parport", 7) == 0)
+#if defined(HAVE_LINUX_PPDEV_H) || defined(HAVE_DEV_PPBUS_PPI_H)
+				else if ((fd = dev_is_parport(ndev)) != -1)
 				{
-					cwdev->free (cwdev);
+					cwdev->free(cwdev);
 					cwdev = &cwdevice_lp;
-					cwdev->desc = actual_device(message + 2);
-					cwdev->init (cwdev);
-				}
-				else if (strncmp (message + 2, "ppi", 3) == 0)
-				{
-					cwdev->free (cwdev);
-					cwdev = &cwdevice_lp;
-					cwdev->desc = message + 2;
-					cwdev->init (cwdev);
+					keydev = cwdev->desc = ndev;
+					cwdev->init(cwdev, fd);
 				}
 #endif
+				else if ((fd = dev_is_null(ndev)) != -1)
+				{
+					cwdev->free(cwdev);
+					cwdev = &cwdevice_null;
+					keydev = cwdev->desc = ndev;
+					cwdev->init(cwdev, fd);
+				}
 				else debug ("Unknown device");
 				break;
 			case '9':	/* base port number */
 				debug ("Obsolete");
 				break;
 			case 'a':	/* PTT keying on or off */
-				if (atoi (message + 2))
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv)
 				{
 					cwdev->ptt (cwdev, ON);
 					debug ("PTT on");
@@ -461,7 +473,9 @@ recv_code (void)
 				}
 				break;
 			case 'b':	/* SSB way */
-				if (atoi (message + 2))
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv)
 				{
 					if (cwdev->ssbway)
 					{
@@ -483,32 +497,41 @@ recv_code (void)
 				}
 				break;
 			case 'c':	/* Tune for a number of seconds */
-				tunesec = atoi (message + 2);
-				if (tunesec <= 10)
-					tune (tunesec);
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv <= 10)
+					tune(lv);
 				break;
 			case 'd':	/* set ptt delay (TOD, Turn On Delay) */
-				ptt_delay = atoi (message + 2);
-				if ((ptt_delay >= 0) && (ptt_delay < 51))	/* only allowed range 0..50 ms */
-					ptt_delay *= 1000;	/* 0 = ptt delay off */
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv >= 0 && lv < 51)
+					ptt_delay = lv * 1000;
 				else
-					ptt_delay = 50000;	/* longest delay (secure) if parameter wrong */
+					ptt_delay = 50000;
 				debug ("PTT delay(TOD): %d ms", ptt_delay / 1000);
-				if (ptt_delay)
-					;
-				else
+				if (ptt_delay == 0)
 				{
 					cwdev->ptt (cwdev, OFF);
 					debug ("PTT off");
 				}
 				break;
 			case 'e':	/* set bandswitch output on parport bits 2(lsb),7,8,9(msb) */
-				bandswitch = atoi (message + 2);
-				if (bandswitch <= 15 && bandswitch >= 0)
-					set_switch (bandswitch);
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv <= 15 && lv >= 0) {
+					bandswitch = lv;
+					set_switch(bandswitch);
+				}
 				break;
 			case 'f':	/* switch console/soundcard */
-				if (!strncmp (message + 2, "c", 1))
+				if (!strncmp (message + 2, "n", 1))
+				{
+					console_sound = 0;
+					soundcard_sound = 0;
+					valid_sdevice = 1;
+				}
+				else if (!strncmp (message + 2, "c", 1))
 				{
 					console_sound = 1;
 					soundcard_sound = 0;
@@ -534,21 +557,22 @@ recv_code (void)
 				}
 				break;
 			case 'g':	/* volume */
-				vol = atoi (message + 2);
-				if (vol <= 100 && vol >= 0)
+				if (get_long(message + 2, &lv))
+					break;
+				if (lv <= 100 && lv >= 0)
 				{
-					morse_volume = vol;
+					morse_volume = lv;
 					cw_set_volume (morse_volume);
 				}
 				break;
-            case 'h':   /* send echo to main program when CW playing is done */
-                memcpy(&reply_sin, &k_sin, sizeof(reply_sin)); /* remember sender */
-                reply_socklen = sin_len;
-                strncpy(reply_data+1, message, sizeof(reply_data) - 2);
-                reply_data[sizeof(reply_data) - 1] = '\0';
-                reply_data[0]='h';
-                if (strlen (message) + 1 <= MAXMORSE - 1) strcat (morsetext, "^");
-                break;
+	    case 'h':   /* send echo to main program when CW playing is done */
+		memcpy(&reply_sin, &k_sin, sizeof(reply_sin)); /* remember sender */
+		reply_socklen = sin_len;
+		strncpy(reply_data+1, message, sizeof(reply_data) - 2);
+		reply_data[sizeof(reply_data) - 1] = '\0';
+		reply_data[0]='h';
+		if (strlen (message) + 1 <= MAXMORSE - 1) strcat (morsetext, "^");
+		break;
 			}
 			return 0;
 		}
@@ -582,15 +606,15 @@ playmorsestring (char *x)
 		}
 		else if (c == '~')
 			cw_set_gap (2); /* 2 dots time additional for the next char */
-        else if (c == '^'){
-            debug("Echo '%s'", reply_data);
-            if (strlen(reply_data)==0) return;
-            sendto(socket_descriptor, reply_data, strlen(reply_data), 0, 
-                (struct sockaddr *)&reply_sin, reply_socklen);
-            strcpy(reply_data,"");
-            break;
+	else if (c == '^'){
+	    debug("Echo '%s'", reply_data);
+	    if (strlen(reply_data)==0) return;
+	    sendto(socket_descriptor, reply_data, strlen(reply_data), 0, 
+		(struct sockaddr *)&reply_sin, reply_socklen);
+	    strcpy(reply_data,"");
+	    break;
   
-        }
+	}
 		else
 		{
 			validchar = 1;
@@ -656,7 +680,8 @@ static void pttevent (int pttstate)
 static void
 parsecommandline (int argc, char *argv[])
 {
-	int p, w;
+	long lv;
+	int p;
 
 	while ((p = getopt (argc, argv, "d:hnp:P:s:t:v:Vw:x:")) != -1)
 	{
@@ -700,117 +725,75 @@ parsecommandline (int argc, char *argv[])
 			printf ("       -w <weight>   ");
 			printf ("Set weighting (-50 ... 50, default = 0)\n");
 			printf ("       -x <sdevice>  ");
-			printf ("Use a different sound device\n                     ");
-			printf ("(c = console (default), s = soundcard, b = both)\n");
+			printf ("Use a different sound device\n		     ");
+			printf ("(c = console (default), s = soundcard, b = both, n = none)\n");
 			exit (0);
 		case 'd':
-			if (!strncmp (optarg, "tty", 3)
-			    && strlen (optarg) < 9)
-				cwdev = &cwdevice_ttys;
-#ifdef HAVE_LINUX_PPDEV_H
-			else if (!strncmp (optarg, "parport", 7)
-				 && strlen (optarg) < 9)
-				cwdev = &cwdevice_lp;
-#endif
-#ifdef HAVE_DEV_PPBUS_PPI_H
-			else if ((strncmp (optarg, "parport", 7) == 0)
-				 && strlen (optarg) < 9)
-				cwdev = &cwdevice_lp;
-			else if ((strncmp (optarg, "ppi", 3) == 0)
-				 && strlen (optarg) < 5)
-				cwdev = &cwdevice_lp;
-#endif
-			else
-			{
-				printf ("%s: Wrong device %s, ", PACKAGE,
-					optarg);
-#if defined (HAVE_LINUX_PPDEV_H)
-				printf ("use one of ttyS0,1,2,3 or parport0,1 etc.\n");
-#elif defined (HAVE_DEV_PPBUS_PPI_H)
-				printf ("use one of ttyd0,1,2,3 or ppi0,1 etc.\n");
-#else
-#ifdef BSD
-				printf ("use one of ttyd0,1,2,3 etc.\n");
-#else
-				printf ("use one of ttyS0,1,2,3 etc.\n");
-#endif
-#endif
-				exit (1);
-			}
-#ifdef HAVE_DEV_PPBUS_PPI_H
-			cwdev->desc = actual_device(optarg);
-#else
-			cwdev->desc = optarg;
-#endif
+			keydev = optarg;
 			break;
 		case 'n':
 			forking = 0;
 			printf ("%s: Not forking...\n", PACKAGE);
 			break;
 		case 'p':
-			if ((port = atoi (optarg)) <= 0)
-			{
-				printf ("%s: Invalid port number - %s\n",
-					PACKAGE, optarg);
-				exit (1);
-			}
-			if (port < 1024)
-			{
-				printf ("%s: Port should be larger than 1023\n", PACKAGE);
-				exit (1);
+			if (get_long(optarg, &lv) || lv < 1024 || lv > 65536) {
+				printf("%s: invalid port number - %s\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
+			port = lv;
 			break;
 #if defined(HAVE_SETPRIORITY) && defined(PRIO_PROCESS)
 		case 'P':
-			priority = atoi (optarg);
-			if (priority < -20 || priority > 20)
-			{
-				printf ("%s: Wrong priority, use a value between -20 and 20\n",
-					PACKAGE);
-				exit (1);
+			if (get_long(optarg, &lv) || lv < -20 || lv > 20) {
+				printf("%s: bad priority %s (should be between -20 and 20 inclusive)\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
-		break;
+			priority = lv;
+			break;
 #endif
 		case 's':
-			morse_speed = atoi (optarg);
-			if (morse_speed < 5 || morse_speed > 60)
-			{
-				printf ("Wrong speed, use a value between 4 and 60\n");
-				exit (1);
+			if (get_long(optarg, &lv) || lv < 4 || lv > 60) {
+				printf("%s: bad speed %s (should be between 4 and 60 inclusive)\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
+			morse_speed = lv;
 			break;
 		case 't':
-			ptt_delay = atoi (optarg);
-			if (ptt_delay < 0 || ptt_delay > 50)
-			{
-				printf ("Wrong PTT delay value, must be between 0 and 50 ms\n");
-				exit (1);
+			if (get_long(optarg, &lv) || lv < 0 || lv > 50) {
+				printf("%s: bad PTT delay value %s (should be between 0 and 50 ms inclusive)\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
-			else
-				ptt_delay *= 1000;
+			ptt_delay = 1000 * lv;
 			break;
 		case 'v':
-			morse_volume = atoi (optarg);
-			if ((morse_volume < 0) || (morse_volume > 100))
-			{
-				printf ("Wrong volume value, use 0 ... 100\n");
-				exit (1);
+			if (get_long(optarg, &lv) || lv < 0 || lv > 100) {
+				printf("%s: bad volume %s (should be between 0 and 100 inclusive)\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
+			morse_volume = lv;
 			break;
 		case 'V':
 			printf ("%s version %s\n", PACKAGE, VERSION);
 			exit (0);
 		case 'w':
-			w = atoi (optarg);
-			if ((w > 50) || (w < -50))
-			{
-				printf ("Wrong weighting value, use -50 ... 50\n");
-				exit (1);
+			if (get_long(optarg, &lv) || lv < -50 || lv > 50) {
+				printf("%s: bad weight %s (should be between -50 and 50 inclusive)\n",
+				    PACKAGE, optarg);
+				exit(1);
 			}
-			cw_set_weighting (w);
+			cw_set_weighting (lv);
 			break;
 		case 'x':
-			if (!strncmp(optarg, "c", 1))
+			if (!strncmp(optarg, "n", 1)) {
+				console_sound = 0;
+				soundcard_sound = 0;
+			}
+			else if (!strncmp(optarg, "c", 1))
 			{
 				console_sound = 1;
 				soundcard_sound = 0;
@@ -827,7 +810,7 @@ parsecommandline (int argc, char *argv[])
 			}
 			else
 			{
-				printf ("Wrong sound device, use c(onsole), s(soundcard) or b(oth)\n");
+				printf ("Wrong sound device, use c(onsole), s(oundcard), b(oth), or n(one)\n");
 				exit (1);
 			}
 			break;
@@ -841,19 +824,21 @@ int
 main (int argc, char *argv[])
 {
 	pid_t pid, sid;
-	int bind_rc, close_rc;
-	long save_file_flags;
+	int fd;
+	int save_file_flags;
 
 #if defined HAVE_LINUX_PPDEV_H
-	cwdev->desc = "parport0";
+	keydev = "parport0";
 #elif defined HAVE_DEV_PPBUS_PPI_H
-	cwdev->desc = "ppi0";
+	keydev = "ppi0";
 #else
- #ifdef BSD
-	cwdev->desc = "ttyd0";
- #else
-	cwdev->desc = "ttyS0";
- #endif
+# if defined(__FreeBSD__)
+	keydev = "ttyd0";
+# elif defined(__OpenBSD__)
+	keydev = "tty00";
+# else
+	keydev = "ttyS0";
+# endif
 #endif
 
 	parsecommandline (argc, argv);
@@ -862,11 +847,29 @@ main (int argc, char *argv[])
 		printf ("You must run this program as root\n");
 		exit (1);
 	}
+
+	if ((fd = dev_is_tty(keydev)) != -1)
+		cwdev = &cwdevice_ttys;
+#if defined (HAVE_LINUX_PPDEV_H) || defined (HAVE_DEV_PPBUS_PPI_H)
+	else if ((fd = dev_is_parport(keydev)) != -1)
+		cwdev = &cwdevice_lp;
+#endif
+	else if ((fd = dev_is_null(keydev)) != -1)
+		cwdev = &cwdevice_null;
+
+	if (cwdev == NULL)
+	{
+		printf("%s: bad keyer device: %s\n", PACKAGE, keydev);
+		exit(1);
+	}
+	cwdev->desc = keydev;
+
 	initmorse ();
 	cw_keying_callback (keyingevent);
 	cw_ptt_callback (pttevent);
 
 	debug ("Device used: %s", cwdev->desc);
+	cwdev->init (cwdev, fd);
 
 	if (forking)
 	{
@@ -893,9 +896,32 @@ main (int argc, char *argv[])
 			exit (1);
 		}
 		umask (0);
-		close (STDIN_FILENO);
-		close (STDOUT_FILENO);
-		close (STDERR_FILENO);
+
+		/* replace stdin/stdout/stderr with /dev/null */
+		if ((fd = open("/dev/null", O_RDWR, 0)) == -1)
+		{
+		        syslog (LOG_ERR, "%s\n", "open null");
+		        exit (1);
+		}
+		if (dup2(fd, STDIN_FILENO) == -1)
+		{
+		        syslog (LOG_ERR, "%s\n", "dup2 stdin");
+		        exit (1);
+		}
+		if (dup2(fd, STDOUT_FILENO) == -1)
+		{
+		        syslog (LOG_ERR, "%s\n", "dup2 stdout");
+		        exit (1);
+		}
+		if (dup2(fd, STDERR_FILENO) == -1)
+		{
+		        syslog (LOG_ERR, "%s\n", "dup2 stderr");
+		        exit (1);
+		}
+		if (fd != STDIN_FILENO &&
+		    fd != STDOUT_FILENO &&
+		    fd != STDERR_FILENO)
+		        (void)close(fd);
 	}
 	else
 	{
@@ -903,8 +929,6 @@ main (int argc, char *argv[])
 		signal (SIGINT, catchint);
 	}
 
-	cwdev->init (cwdev);
-
 	bzero (&k_sin, sizeof (k_sin));
 	k_sin.sin_family = AF_INET;
 	k_sin.sin_addr.s_addr = htonl (INADDR_ANY);
@@ -918,16 +942,19 @@ main (int argc, char *argv[])
 		exit (1);
 	}
 
-	bind_rc =
-		bind (socket_descriptor, (struct sockaddr *) &k_sin,
-		      sizeof (k_sin));
-	if (bind_rc == -1)
+	if (bind(socket_descriptor, (struct sockaddr *) &k_sin,
+		sizeof (k_sin)) == -1)
 	{
 		errmsg ("Bind");
 		exit (1);
 	}
 
 	save_file_flags = fcntl (socket_descriptor, F_GETFL);
+	if (save_file_flags == -1)
+	{
+		errmsg("Trying get flags");
+		exit(1);
+	}
 	save_file_flags |= O_NONBLOCK;
 
 	if (fcntl (socket_descriptor, F_SETFL, save_file_flags) == -1)
@@ -969,40 +996,10 @@ main (int argc, char *argv[])
 	}
 
 	cwdev->free (cwdev);
-	close_rc = close (socket_descriptor);
-	if (close_rc == -1)
+	if (close (socket_descriptor) == -1)
 	{
 		errmsg ("Close socket");
 		exit (1);
 	}
 	exit (0);
 }
-
-#ifdef HAVE_DEV_PPBUS_PPI_H
-/*
- * inputs	- parport linux name
- * output	- pointer to *static* char with BSD equivalent name
- * side effects - none
- */
-
-static char *
-actual_device(const char *in)
-{
-	static char actual_name[MAX_DEVICE];
-	/*
-	 * parportN
-	 * 01234567
-	 */
-	if (strncmp(in, "parport", 7) == 0)
-	{
-		/* Note no possible overflow here, snprintf not needed. */
-		sprintf(actual_name, "ppi%c", in[7]);
-	}
-	else
-	{
-		strncpy(actual_name, in, MAX_DEVICE-2);
-		actual_name[MAX_DEVICE-1] = '\0';
-	}
-	return(actual_name);
-}
-#endif
